@@ -15,18 +15,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import collections
-
 from oslo_config import cfg
 from oslo_log import log as logging
 
+from networking_bagpipe.bagpipe_bgp.common import config
+from networking_bagpipe.bagpipe_bgp.common import dataplane_utils
 from networking_bagpipe.bagpipe_bgp.common import log_decorator
 from networking_bagpipe.bagpipe_bgp import constants as consts
 from networking_bagpipe.bagpipe_bgp.engine import exa
 from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
 from networking_bagpipe.bagpipe_bgp.vpn import evpn
 
-from neutron.agent.common import ovs_lib
 from neutron.plugins.ml2.drivers.openvswitch.agent.common import \
     constants as ovs_const
 from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.ovs_ofctl import \
@@ -34,7 +33,6 @@ from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.ovs_ofctl import \
 from neutron.plugins.ml2.drivers.openvswitch.agent import ovs_neutron_agent
 
 from neutron_lib import constants as n_consts
-from neutron_lib import exceptions
 
 LOG = logging.getLogger(__name__)
 
@@ -42,89 +40,6 @@ FLOOD = "flood"
 FLOW_PRIORITY = 5
 
 
-# largely copied from networking_sfc.services.sfc.common.ovs_ext_lib
-class OVSBridgeWithGroups(object):
-
-    def __init__(self, ovs_bridge):
-        self.bridge = ovs_bridge
-
-        # OpenFlow 1.1 is needed to manipulate groups
-        self.bridge.use_at_least_protocol(ovs_const.OPENFLOW11)
-
-    # proxy most methods to self.bridge
-    def __getattr__(self, name):
-        return getattr(self.bridge, name)
-
-    def do_action_groups(self, action, kwargs_list):
-        group_strs = [_build_group_expr_str(kw, action) for kw in kwargs_list]
-        if action == 'add' or action == 'del':
-            cmd = '%s-groups' % action
-        elif action == 'mod':
-            cmd = '%s-group' % action
-        else:
-            msg = _("Action is illegal")
-            raise exceptions.InvalidInput(error_message=msg)
-        self.run_ofctl(cmd, ['--may-create', '-'], '\n'.join(group_strs))
-
-    @log_decorator.log_info
-    def add_group(self, **kwargs):
-        self.do_action_groups('add', [kwargs])
-
-    @log_decorator.log_info
-    def mod_group(self, **kwargs):
-        self.do_action_groups('mod', [kwargs])
-
-    @log_decorator.log_info
-    def delete_group(self, **kwargs):
-        self.do_action_groups('del', [kwargs])
-
-    def dump_group_for_id(self, group_id):
-        retval = None
-        group_str = "%d" % group_id
-        group = self.run_ofctl("dump-groups", [group_str])
-        if group:
-            retval = '\n'.join(item for item in group.splitlines()
-                               if ovs_lib.is_a_flow_line(item))
-        return retval
-
-    def get_bridge_ports(self):
-        port_name_list = self.bridge.get_port_name_list()
-        of_portno_list = list()
-        for port_name in port_name_list:
-            of_portno_list.append(self.bridge.get_port_ofport(port_name))
-        return of_portno_list
-
-
-def _build_group_expr_str(group_dict, cmd):
-    group_expr_arr = []
-    buckets = None
-    group_id = None
-
-    if cmd != 'del':
-        if "group_id" not in group_dict:
-            msg = _("Must specify one group Id on group addition"
-                    " or modification")
-            raise exceptions.InvalidInput(error_message=msg)
-        group_id = "group_id=%s" % group_dict.pop('group_id')
-
-        if "buckets" not in group_dict:
-            msg = _("Must specify one or more buckets on group addition"
-                    " or modification")
-            raise exceptions.InvalidInput(error_message=msg)
-        buckets = "%s" % group_dict.pop('buckets')
-
-    if group_id:
-        group_expr_arr.append(group_id)
-
-    for key, value in group_dict.items():
-        group_expr_arr.append("%s=%s" % (key, value))
-
-    if buckets:
-        group_expr_arr.append(buckets)
-
-    return ','.join(group_expr_arr)
-
-
 class OVSEVIDataplane(evpn.VPNInstanceDataplane):
 
     def __init__(self, *args, **kwargs):
@@ -134,7 +49,10 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         # OpenFlow 1.3 is needed for mod_vlan_vid
         self.bridge.use_at_least_protocol(ovs_const.OPENFLOW13)
 
-        self.tunnel_mgr = self.driver.tunnel_mgr
+        self.tunnel_mgr = dataplane_utils.SharedObjectLifecycleManagerProxy(
+            self.driver.tunnel_mgr,
+            self.instance_id
+        )
         self.flooding_ports = set()
         self.vlan = None
         self.local_ip = self.driver.get_local_address()
@@ -176,9 +94,6 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         return "load:0->NXM_OF_IN_PORT[],set_tunnel:%d,resubmit(,%s)" % (
             vni, ovs_const.VXLAN_TUN_TO_LV)
 
-    def _tun_mgr_handle(self, vni, mac):
-        return (self.instance_id, vni, mac)
-
     @log_decorator.log_info
     def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, vni, nlri,
                                             encaps):
@@ -191,8 +106,7 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         if remote_pe == self.local_ip:
             actions = self._local_vni_actions(vni)
         else:
-            port = self.tunnel_mgr.tunnel_for_remote_ip(
-                remote_pe, self._tun_mgr_handle(vni, mac))
+            port, _ = self.tunnel_mgr.get_object(remote_pe, (vni, mac))
             actions = "set_tunnel:%d,output:%s" % (vni, port)
         self.bridge.add_flow(table=ovs_const.UCAST_TO_TUN,
                              priority=FLOW_PRIORITY,
@@ -213,8 +127,7 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         self.bridge.delete_unicast_to_tun(self.vlan, mac)
 
         if remote_pe != self.local_ip:
-            self.tunnel_mgr.free_tunnel(remote_pe,
-                                        self._tun_mgr_handle(vni, mac))
+            self.tunnel_mgr.free_object(remote_pe, (vni, mac))
 
         # cleanup ARP responder
         if ip:
@@ -226,8 +139,7 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         if remote_pe == self.local_ip:
             port = "local"
         else:
-            port = self.tunnel_mgr.tunnel_for_remote_ip(
-                remote_pe, self._tun_mgr_handle(vni, FLOOD))
+            port, _ = self.tunnel_mgr.get_object(remote_pe, (vni, FLOOD))
         self.flooding_ports.add((port, vni))
 
         self._update_flooding_buckets()
@@ -238,15 +150,15 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         if remote_pe == self.local_ip:
             port = "local"
         else:
-            port = self.tunnel_mgr.tunnel_for_remote_ip(remote_pe)
+            port = self.tunnel_mgr.find_object(remote_pe)
 
-        self.flooding_ports.remove((port, vni))
+        if port:
+            self.flooding_ports.remove((port, vni))
 
-        self._update_flooding_buckets()
+            self._update_flooding_buckets()
 
-        if remote_pe != self.local_ip:
-            self.tunnel_mgr.free_tunnel(
-                remote_pe, self._tun_mgr_handle(vni, FLOOD))
+            if remote_pe != self.local_ip:
+                self.tunnel_mgr.free_object(remote_pe, (vni, FLOOD))
 
     def _update_flooding_buckets(self):
         buckets = []
@@ -286,61 +198,31 @@ class OVSEVIDataplane(evpn.VPNInstanceDataplane):
         }
 
 
-class TunnelManager(object):
+class TunnelManager(dataplane_utils.ObjectLifecycleManager):
 
     def __init__(self, bridge, local_ip):
-        self.tunnels = dict()
+        super(TunnelManager, self).__init__()
+
         self.bridge = bridge
         self.local_ip = local_ip
-        self.tunnel_used_for = collections.defaultdict(set)
 
     @log_decorator.log_info
-    def tunnel_for_remote_ip(self, remote_ip, use=None):
-        tunnel = self.tunnels.get(remote_ip)
-        if tunnel:
-            LOG.debug("existing tunnel for %s: %s", remote_ip, tunnel)
-        else:
-            if not use:
-                raise Exception("non-existing tunnel for %s", remote_ip)
+    def create_object(self, remote_ip, *args, **kwargs):
+        port_name = ovs_neutron_agent.OVSNeutronAgent.get_tunnel_name(
+            n_consts.TYPE_VXLAN, self.local_ip, remote_ip)
+        tunnel = self.bridge.add_tunnel_port(port_name,
+                                             remote_ip,
+                                             self.local_ip,
+                                             n_consts.TYPE_VXLAN)
 
-            port_name = ovs_neutron_agent.OVSNeutronAgent.get_tunnel_name(
-                n_consts.TYPE_VXLAN, self.local_ip, remote_ip)
-            tunnel = self.bridge.add_tunnel_port(port_name,
-                                                 remote_ip,
-                                                 self.local_ip,
-                                                 n_consts.TYPE_VXLAN)
-
-            self.bridge.setup_tunnel_port(n_consts.TYPE_VXLAN, tunnel)
-            self.tunnels[remote_ip] = tunnel
-            LOG.debug("tunnel for %s: %s (%s)", remote_ip, port_name, tunnel)
-
-        if use:
-            self.tunnel_used_for[remote_ip].add(use)
+        self.bridge.setup_tunnel_port(n_consts.TYPE_VXLAN, tunnel)
+        LOG.debug("tunnel for %s: %s (%s)", remote_ip, port_name, tunnel)
 
         return tunnel
 
     @log_decorator.log_info
-    def free_tunnel(self, remote_ip, use):
-        if remote_ip not in self.tunnel_used_for:
-            LOG.debug("no tunnel to free for %s", remote_ip)
-            return
-
-        self.tunnel_used_for[remote_ip].discard(use)
-        if not self.tunnel_used_for[remote_ip]:
-            tunnel = self.tunnels[remote_ip]
-
-            LOG.debug("%s was last user for %s, clearing port", use,
-                      remote_ip)
-            self.bridge.delete_port(tunnel)
-
-            del self.tunnels[remote_ip]
-            del self.tunnel_used_for[remote_ip]
-        else:
-            LOG.debug("remaining users for tunnel %s: %s", remote_ip,
-                      self.tunnel_used_for[remote_ip])
-
-    def infos(self):
-        return self.tunnels
+    def delete_object(self, tunnel):
+        self.bridge.delete_port(tunnel)
 
 
 class OVSDataplaneDriver(dp_drivers.DataplaneDriver):
@@ -359,7 +241,10 @@ class OVSDataplaneDriver(dp_drivers.DataplaneDriver):
 
     def __init__(self, *args, **kwargs):
         super(OVSDataplaneDriver, self).__init__(*args, **kwargs)
-        self.bridge = OVSBridgeWithGroups(
+
+        config.set_default_root_helper()
+
+        self.bridge = dataplane_utils.OVSBridgeWithGroups(
             br_tun.OVSTunnelBridge(self.config.ovs_bridge)
         )
         self.tunnel_mgr = TunnelManager(self.bridge,
